@@ -43,38 +43,39 @@
 #include <X11/keysym.h>
 #include <X11/Xatom.h>
 
+#include <gdesktop-enums.h>
+
 #include <X11/extensions/XInput.h>
 #include <X11/extensions/XIproto.h>
 
+#include "gnome-settings-bus.h"
 #include "gnome-settings-profile.h"
 #include "gsd-mouse-manager.h"
 #include "gsd-input-helper.h"
 #include "gsd-enums.h"
+#include "gsd-settings-migrate.h"
 
 #define GSD_MOUSE_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSD_TYPE_MOUSE_MANAGER, GsdMouseManagerPrivate))
 
-#define SETTINGS_MOUSE_DIR         "org.gnome.settings-daemon.peripherals.mouse"
-#define SETTINGS_TOUCHPAD_DIR      "org.gnome.settings-daemon.peripherals.touchpad"
-#define SETTINGS_TRACKBALL_DIR     "org.gnome.settings-daemon.peripherals.trackball"
+#define GSD_SETTINGS_MOUSE_SCHEMA  "org.gnome.settings-daemon.peripherals.mouse"
+#define GSETTINGS_MOUSE_SCHEMA     "org.gnome.desktop.peripherals.mouse"
+#define GSETTINGS_TOUCHPAD_SCHEMA  "org.gnome.desktop.peripherals.touchpad"
+#define GSETTINGS_TRACKBALL_SCHEMA "org.gnome.desktop.peripherals.trackball"
 
 /* Keys for both touchpad and mouse */
 #define KEY_LEFT_HANDED         "left-handed"                /* a boolean for mouse, an enum for touchpad */
-#define KEY_MOTION_ACCELERATION "motion-acceleration"
-#define KEY_MOTION_THRESHOLD    "motion-threshold"
+#define KEY_SPEED               "speed"
 
 /* Touchpad settings */
-#define KEY_TOUCHPAD_DISABLE_W_TYPING    "disable-while-typing"
-#define KEY_PAD_HORIZ_SCROLL             "horiz-scroll-enabled"
 #define KEY_SCROLL_METHOD                "scroll-method"
 #define KEY_TAP_TO_CLICK                 "tap-to-click"
-#define KEY_TOUCHPAD_ENABLED             "touchpad-enabled"
+#define KEY_SEND_EVENTS                  "send-events"
 #define KEY_NATURAL_SCROLL_ENABLED       "natural-scroll"
 
 /* Mouse settings */
 #define KEY_LOCATE_POINTER               "locate-pointer"
 #define KEY_DWELL_CLICK_ENABLED          "dwell-click-enabled"
 #define KEY_SECONDARY_CLICK_ENABLED      "secondary-click-enabled"
-#define KEY_MIDDLE_BUTTON_EMULATION      "middle-button-enabled"
 
 /* Trackball settings */
 #define KEY_SCROLL_WHEEL_BUTTON          "scroll-wheel-emulation-button"
@@ -86,6 +87,7 @@ struct GsdMouseManagerPrivate
         GSettings *mouse_settings;
         GSettings *mouse_a11y_settings;
         GSettings *trackball_settings;
+        GSettings *gsd_mouse_settings;
         GdkDeviceManager *device_manager;
         guint device_added_id;
         guint device_removed_id;
@@ -104,6 +106,8 @@ static void     gsd_mouse_manager_finalize    (GObject             *object);
 static void     set_tap_to_click              (GdkDevice           *device,
                                                gboolean             state,
                                                gboolean             left_handed);
+static void     ensure_touchpad_active        (GsdMouseManager     *manager);
+
 
 G_DEFINE_TYPE (GsdMouseManager, gsd_mouse_manager, G_TYPE_OBJECT)
 
@@ -357,6 +361,9 @@ set_left_handed (GsdMouseManager *manager,
         if (!xinput_device_has_buttons (device))
                 return;
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         xdevice = open_gdk_device (device);
         if (xdevice == NULL)
                 return;
@@ -420,8 +427,12 @@ set_motion (GsdMouseManager *manager,
         gfloat motion_acceleration;
         int motion_threshold;
         GSettings *settings;
+        gdouble speed;
         guint i;
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         xdevice = open_gdk_device (device);
         if (xdevice == NULL)
                 return;
@@ -433,8 +444,11 @@ set_motion (GsdMouseManager *manager,
         else
                 settings = manager->priv->mouse_settings;
 
-        /* Calculate acceleration */
-        motion_acceleration = g_settings_get_double (settings, KEY_MOTION_ACCELERATION);
+        speed = g_settings_get_double (settings, KEY_SPEED);
+
+        /* Calculate acceleration and threshold */
+        motion_acceleration = (speed + 1) * 5; /* speed is [-1..1], map to [0..10] */
+        motion_threshold = CLAMP (10 - floor (motion_acceleration), 1, 10);
 
         if (motion_acceleration >= 1.0) {
                 /* we want to get the acceleration, with a resolution of 0.5
@@ -461,9 +475,6 @@ set_motion (GsdMouseManager *manager,
                 denominator = -1;
         }
 
-        /* And threshold */
-        motion_threshold = g_settings_get_int (settings, KEY_MOTION_THRESHOLD);
-
         gdk_error_trap_push ();
 
         /* Get the list of feedbacks for the device */
@@ -504,53 +515,6 @@ set_motion (GsdMouseManager *manager,
         xdevice_close (xdevice);
 }
 
-static void
-set_middle_button (GsdMouseManager *manager,
-                   GdkDevice       *device,
-                   gboolean         middle_button)
-{
-        Atom prop;
-        XDevice *xdevice;
-        Atom type;
-        int format;
-        unsigned long nitems, bytes_after;
-        unsigned char *data;
-        int rc;
-
-        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
-                            "Evdev Middle Button Emulation", True);
-
-        if (!prop) /* no evdev devices */
-                return;
-
-        xdevice = open_gdk_device (device);
-        if (xdevice == NULL)
-                return;
-
-	g_debug ("setting middle button on %s", gdk_device_get_name (device));
-
-        gdk_error_trap_push ();
-
-        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
-                            xdevice, prop, 0, 1, False, XA_INTEGER, &type, &format,
-                            &nitems, &bytes_after, &data);
-
-        if (rc == Success && format == 8 && type == XA_INTEGER && nitems == 1) {
-                data[0] = middle_button ? 1 : 0;
-
-                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
-                                       xdevice, prop, type, format, PropModeReplace, data, nitems);
-        }
-
-        if (gdk_error_trap_pop ())
-                g_warning ("Error in setting middle button emulation on \"%s\"", gdk_device_get_name (device));
-
-        if (rc == Success)
-                XFree (data);
-
-        xdevice_close (xdevice);
-}
-
 /* Ensure that syndaemon dies together with us, to avoid running several of
  * them */
 static void
@@ -619,7 +583,6 @@ set_disable_w_typing (GsdMouseManager *manager, gboolean state)
 
                 if (error) {
                         g_warning ("Failed to launch syndaemon: %s", error->message);
-                        g_settings_set_boolean (manager->priv->touchpad_settings, KEY_TOUCHPAD_DISABLE_W_TYPING, FALSE);
                         g_error_free (error);
                 } else {
                         g_child_watch_add (manager->priv->syndaemon_pid, syndaemon_died, manager);
@@ -646,6 +609,9 @@ set_tap_to_click (GdkDevice *device,
         unsigned char* data;
         Atom prop, type;
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Tap Action", False);
         if (!prop)
                 return;
@@ -695,6 +661,9 @@ set_horiz_scroll (GdkDevice *device,
         unsigned long nitems, bytes_after;
         unsigned char *data;
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         prop_edge = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Edge Scrolling", False);
         prop_twofinger = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Two-Finger Scrolling", False);
 
@@ -761,6 +730,9 @@ set_scroll_method (GsdMouseManager         *manager,
         unsigned long nitems, bytes_after;
         unsigned char *data;
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Capabilities", True);
         prop_edge = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Edge Scrolling", False);
         prop_twofinger = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Two-Finger Scrolling", False);
@@ -835,6 +807,9 @@ set_touchpad_disabled (GdkDevice *device)
         int id;
         XDevice *xdevice;
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         g_object_get (G_OBJECT (device), "device-id", &id, NULL);
 
         g_debug ("Trying to set device disabled for \"%s\" (%d)", gdk_device_get_name (device), id);
@@ -861,6 +836,9 @@ set_touchpad_enabled (int id)
 {
         XDevice *xdevice;
 
+        if (xdevice_is_libinput (id))
+                return;
+
         g_debug ("Trying to set device enabled for %d", id);
 
         gdk_error_trap_push ();
@@ -902,7 +880,7 @@ set_locate_pointer (GsdMouseManager *manager,
                 manager->priv->locate_pointer_spawned = (error == NULL);
 
                 if (error) {
-                        g_settings_set_boolean (manager->priv->mouse_settings, KEY_LOCATE_POINTER, FALSE);
+                        g_settings_set_boolean (manager->priv->gsd_mouse_settings, KEY_LOCATE_POINTER, FALSE);
                         g_error_free (error);
                 }
 
@@ -984,6 +962,9 @@ set_natural_scroll (GsdMouseManager *manager,
                 return;
         }
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         g_debug ("Trying to set %s for \"%s\"",
                  natural_scroll ? "natural (reverse) scroll" : "normal scroll",
                  gdk_device_get_name (device));
@@ -1040,6 +1021,9 @@ set_scroll_wheel_button (GsdMouseManager *manager,
         if (!device_is_trackball (device))
                 return;
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         xdevice = open_gdk_device (device);
         if (xdevice == NULL)
                 return;
@@ -1101,24 +1085,42 @@ set_scroll_wheel_button (GsdMouseManager *manager,
         xdevice_close (xdevice);
 }
 
+static gboolean
+get_touchpad_enabled (GsdMouseManager *manager)
+{
+        GDesktopDeviceSendEvents send_events;
+
+        send_events = g_settings_get_enum (manager->priv->touchpad_settings, KEY_SEND_EVENTS);
+
+        if (send_events == G_DESKTOP_DEVICE_SEND_EVENTS_DISABLED_ON_EXTERNAL_MOUSE) {
+                /* FIXME: mouse_is_present() also finds internal ones... */
+                return (!mouse_is_present () && !trackball_is_present ());
+
+        }
+
+        return send_events == G_DESKTOP_DEVICE_SEND_EVENTS_ENABLED ? TRUE : FALSE;
+}
+
 static void
 set_mouse_settings (GsdMouseManager *manager,
                     GdkDevice       *device)
 {
         gboolean mouse_left_handed, touchpad_left_handed;
 
+        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                return;
+
         mouse_left_handed = g_settings_get_boolean (manager->priv->mouse_settings, KEY_LEFT_HANDED);
         touchpad_left_handed = get_touchpad_handedness (manager, mouse_left_handed);
         set_left_handed (manager, device, mouse_left_handed, touchpad_left_handed);
 
         set_motion (manager, device);
-        set_middle_button (manager, device, g_settings_get_boolean (manager->priv->mouse_settings, KEY_MIDDLE_BUTTON_EMULATION));
 
         set_tap_to_click (device, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_TAP_TO_CLICK), touchpad_left_handed);
         set_scroll_method (manager, device, g_settings_get_enum (manager->priv->touchpad_settings, KEY_SCROLL_METHOD));
-        set_horiz_scroll (device, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_PAD_HORIZ_SCROLL));
+        set_horiz_scroll (device, TRUE);
         set_natural_scroll (manager, device, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_NATURAL_SCROLL_ENABLED));
-        if (g_settings_get_boolean (manager->priv->touchpad_settings, KEY_TOUCHPAD_ENABLED) == FALSE)
+        if (!get_touchpad_enabled (manager))
                 set_touchpad_disabled (device);
 
         set_scroll_wheel_button (manager, device);
@@ -1150,15 +1152,15 @@ mouse_callback (GSettings       *settings,
                 if (device_is_ignored (manager, device))
                         continue;
 
+                if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                        continue;
+
                 if (g_str_equal (key, KEY_LEFT_HANDED)) {
                         gboolean mouse_left_handed;
                         mouse_left_handed = g_settings_get_boolean (settings, KEY_LEFT_HANDED);
                         set_left_handed (manager, device, mouse_left_handed, get_touchpad_handedness (manager, mouse_left_handed));
-                } else if (g_str_equal (key, KEY_MOTION_ACCELERATION) ||
-                           g_str_equal (key, KEY_MOTION_THRESHOLD)) {
+                } else if (g_str_equal (key, KEY_SPEED)) {
                         set_motion (manager, device);
-                } else if (g_str_equal (key, KEY_MIDDLE_BUTTON_EMULATION)) {
-                        set_middle_button (manager, device, g_settings_get_boolean (settings, KEY_MIDDLE_BUTTON_EMULATION));
                 }
         }
         g_list_free (devices);
@@ -1171,11 +1173,6 @@ touchpad_callback (GSettings       *settings,
 {
         GList *devices, *l;
 
-        if (g_str_equal (key, KEY_TOUCHPAD_DISABLE_W_TYPING)) {
-                set_disable_w_typing (manager, g_settings_get_boolean (manager->priv->touchpad_settings, key));
-                return;
-        }
-
         devices = gdk_device_manager_list_devices (manager->priv->device_manager, GDK_DEVICE_TYPE_SLAVE);
 
         for (l = devices; l != NULL; l = l->next) {
@@ -1184,6 +1181,9 @@ touchpad_callback (GSettings       *settings,
                 if (device_is_ignored (manager, device))
                         continue;
 
+                if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                        continue;
+
                 if (g_str_equal (key, KEY_TAP_TO_CLICK)) {
                         gboolean mouse_left_handed;
                         mouse_left_handed = g_settings_get_boolean (manager->priv->mouse_settings, KEY_LEFT_HANDED);
@@ -1191,16 +1191,13 @@ touchpad_callback (GSettings       *settings,
                                           get_touchpad_handedness (manager, mouse_left_handed));
                 } else if (g_str_equal (key, KEY_SCROLL_METHOD)) {
                         set_scroll_method (manager, device, g_settings_get_enum (settings, key));
-                        set_horiz_scroll (device, g_settings_get_boolean (settings, KEY_PAD_HORIZ_SCROLL));
-                } else if (g_str_equal (key, KEY_PAD_HORIZ_SCROLL)) {
-                        set_horiz_scroll (device, g_settings_get_boolean (settings, key));
-                } else if (g_str_equal (key, KEY_TOUCHPAD_ENABLED)) {
-                        if (g_settings_get_boolean (settings, key) == FALSE)
+                        set_horiz_scroll (device, TRUE);
+                } else if (g_str_equal (key, KEY_SEND_EVENTS)) {
+                        if (!get_touchpad_enabled (manager))
                                 set_touchpad_disabled (device);
                         else
                                 set_touchpad_enabled (gdk_x11_device_get_id (device));
-                } else if (g_str_equal (key, KEY_MOTION_ACCELERATION) ||
-                           g_str_equal (key, KEY_MOTION_THRESHOLD)) {
+                } else if (g_str_equal (key, KEY_SPEED)) {
                         set_motion (manager, device);
                 } else if (g_str_equal (key, KEY_LEFT_HANDED)) {
                         gboolean mouse_left_handed;
@@ -1212,16 +1209,8 @@ touchpad_callback (GSettings       *settings,
         }
         g_list_free (devices);
 
-        if (g_str_equal (key, KEY_TOUCHPAD_ENABLED) &&
-            g_settings_get_boolean (settings, key)) {
-                devices = get_disabled_devices (manager->priv->device_manager);
-                for (l = devices; l != NULL; l = l->next) {
-                        int device_id;
-
-                        device_id = GPOINTER_TO_INT (l->data);
-                        set_touchpad_enabled (device_id);
-                }
-                g_list_free (devices);
+        if (g_str_equal (key, KEY_SEND_EVENTS)) {
+                ensure_touchpad_active (manager);
         }
 }
 
@@ -1240,6 +1229,9 @@ trackball_callback (GSettings       *settings,
                 if (device_is_ignored (manager, device))
                         continue;
 
+                if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                        return;
+
                 set_scroll_wheel_button (manager, device);
         }
         g_list_free (devices);
@@ -1249,8 +1241,36 @@ trackball_callback (GSettings       *settings,
 static void
 ensure_touchpad_active (GsdMouseManager *manager)
 {
-        if (mouse_is_present () == FALSE && touchscreen_is_present () == FALSE && trackball_is_present () == FALSE && touchpad_is_present ())
-                g_settings_set_boolean (manager->priv->touchpad_settings, KEY_TOUCHPAD_ENABLED, TRUE);
+        GList *devices, *l;
+
+        if (get_touchpad_enabled (manager)) {
+                devices = get_disabled_devices (manager->priv->device_manager);
+                for (l = devices; l != NULL; l = l->next) {
+                        int device_id;
+
+                        device_id = GPOINTER_TO_INT (l->data);
+                        set_touchpad_enabled (device_id);
+                }
+                g_list_free (devices);
+        } else {
+                devices = gdk_device_manager_list_devices (manager->priv->device_manager,
+                                                           GDK_DEVICE_TYPE_SLAVE);
+
+                for (l = devices; l != NULL; l = l->next) {
+                        GdkDevice *device = l->data;
+
+                        if (device_is_ignored (manager, device))
+                                continue;
+                        if (xdevice_is_libinput (gdk_x11_device_get_id (device)))
+                                continue;
+                        if (gdk_device_get_source (device) != GDK_SOURCE_TOUCHPAD)
+                                continue;
+
+                        set_touchpad_disabled (device);
+                }
+
+                g_list_free (devices);
+        }
 }
 
 static void
@@ -1269,7 +1289,10 @@ device_added_cb (GdkDeviceManager *device_manager,
                 }
 
                 /* If a touchpad was to appear... */
-                set_disable_w_typing (manager, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_TOUCHPAD_DISABLE_W_TYPING));
+                set_disable_w_typing (manager, TRUE);
+
+                /* If a mouse was to appear... */
+                ensure_touchpad_active (manager);
         }
 }
 
@@ -1290,7 +1313,7 @@ device_removed_cb (GdkDeviceManager *device_manager,
                 run_custom_command (device, COMMAND_DEVICE_REMOVED);
 
                 /* If a touchpad was to disappear... */
-                set_disable_w_typing (manager, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_TOUCHPAD_DISABLE_W_TYPING));
+                set_disable_w_typing (manager, TRUE);
 
                 ensure_touchpad_active (manager);
         }
@@ -1326,29 +1349,33 @@ gsd_mouse_manager_idle_cb (GsdMouseManager *manager)
 
         set_devicepresence_handler (manager);
 
-        manager->priv->mouse_settings = g_settings_new (SETTINGS_MOUSE_DIR);
-        g_signal_connect (manager->priv->mouse_settings, "changed",
+        manager->priv->gsd_mouse_settings = g_settings_new (GSD_SETTINGS_MOUSE_SCHEMA);
+        g_signal_connect (manager->priv->gsd_mouse_settings, "changed",
                           G_CALLBACK (mouse_callback), manager);
 
         manager->priv->mouse_a11y_settings = g_settings_new ("org.gnome.desktop.a11y.mouse");
         g_signal_connect (manager->priv->mouse_a11y_settings, "changed",
                           G_CALLBACK (mouse_callback), manager);
 
-        manager->priv->touchpad_settings = g_settings_new (SETTINGS_TOUCHPAD_DIR);
+        manager->priv->mouse_settings = g_settings_new (GSETTINGS_MOUSE_SCHEMA);
+        g_signal_connect (manager->priv->mouse_settings, "changed",
+                          G_CALLBACK (mouse_callback), manager);
+
+        manager->priv->touchpad_settings = g_settings_new (GSETTINGS_TOUCHPAD_SCHEMA);
         g_signal_connect (manager->priv->touchpad_settings, "changed",
                           G_CALLBACK (touchpad_callback), manager);
 
-        manager->priv->trackball_settings = g_settings_new (SETTINGS_TRACKBALL_DIR);
+        manager->priv->trackball_settings = g_settings_new (GSETTINGS_TRACKBALL_SCHEMA);
         g_signal_connect (manager->priv->trackball_settings, "changed",
                           G_CALLBACK (trackball_callback), manager);
 
         manager->priv->syndaemon_spawned = FALSE;
 
-        set_locate_pointer (manager, g_settings_get_boolean (manager->priv->mouse_settings, KEY_LOCATE_POINTER));
+        set_locate_pointer (manager, g_settings_get_boolean (manager->priv->gsd_mouse_settings, KEY_LOCATE_POINTER));
         set_mousetweaks_daemon (manager,
                                 g_settings_get_boolean (manager->priv->mouse_a11y_settings, KEY_DWELL_CLICK_ENABLED),
                                 g_settings_get_boolean (manager->priv->mouse_a11y_settings, KEY_SECONDARY_CLICK_ENABLED));
-        set_disable_w_typing (manager, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_TOUCHPAD_DISABLE_W_TYPING));
+        set_disable_w_typing (manager, TRUE);
 
         devices = gdk_device_manager_list_devices (manager->priv->device_manager, GDK_DEVICE_TYPE_SLAVE);
         for (l = devices; l != NULL; l = l->next) {
@@ -1370,17 +1397,6 @@ gsd_mouse_manager_idle_cb (GsdMouseManager *manager)
 
         ensure_touchpad_active (manager);
 
-        if (g_settings_get_boolean (manager->priv->touchpad_settings, KEY_TOUCHPAD_ENABLED)) {
-                devices = get_disabled_devices (manager->priv->device_manager);
-                for (l = devices; l != NULL; l = l->next) {
-                        int device_id;
-
-                        device_id = GPOINTER_TO_INT (l->data);
-                        set_touchpad_enabled (device_id);
-                }
-                g_list_free (devices);
-        }
-
         gnome_settings_profile_end (NULL);
 
         manager->priv->start_idle_id = 0;
@@ -1399,6 +1415,9 @@ gsd_mouse_manager_start (GsdMouseManager *manager,
                 return TRUE;
         }
 
+        if (gnome_settings_is_wayland ())
+                return TRUE;
+
         manager->priv->start_idle_id = g_idle_add ((GSourceFunc) gsd_mouse_manager_idle_cb, manager);
         g_source_set_name_by_id (manager->priv->start_idle_id, "[gnome-settings-daemon] gsd_mouse_manager_idle_cb");
 
@@ -1453,12 +1472,81 @@ gsd_mouse_manager_finalize (GObject *object)
         G_OBJECT_CLASS (gsd_mouse_manager_parent_class)->finalize (object);
 }
 
+static GVariant *
+map_speed (GVariant *variant)
+{
+        gdouble value;
+
+        value = g_variant_get_double (variant);
+
+        /* Remap from [0..10] to [-1..1] */
+        value = (value / 5) - 1;
+
+        return g_variant_new_double (value);
+}
+
+static GVariant *
+map_send_events (GVariant *variant)
+{
+        gboolean enabled;
+
+        enabled = g_variant_get_boolean (variant);
+
+        if (enabled) {
+                return g_variant_new_string ("enabled");
+        } else {
+                return g_variant_new_string ("disabled");
+        }
+}
+
+static void
+migrate_mouse_settings (void)
+{
+        GsdSettingsMigrateEntry trackball_entries[] = {
+                { "scroll-wheel-emulation-button", "scroll-wheel-emulation-button", NULL }
+        };
+        GsdSettingsMigrateEntry mouse_entries[] = {
+                { "left-handed",           "left-handed", NULL },
+                { "motion-acceleration",   "speed",       map_speed },
+                { "motion-threshold",      NULL,          NULL },
+                { "middle-button-enabled", NULL,          NULL },
+        };
+        GsdSettingsMigrateEntry touchpad_entries[] = {
+                { "disable-while-typing", NULL,             NULL },
+                { "horiz-scroll-enabled", NULL,             NULL },
+                { "scroll-method",        "scroll-method",  NULL },
+                { "tap-to-click",         "tap-to-click",   NULL },
+                { "touchpad-enabled",     "send-events",    map_send_events },
+                { "left-handed",          "left-handed",    NULL },
+                { "motion-acceleration",  "speed",          map_speed },
+                { "motion-threshold",     NULL,             NULL },
+                { "natural-scroll",       "natural-scroll", NULL }
+        };
+
+        gsd_settings_migrate_check ("org.gnome.settings-daemon.peripherals.trackball.deprecated",
+                                    "/org/gnome/settings-daemon/peripherals/trackball/",
+                                    "org.gnome.desktop.peripherals.trackball",
+                                    "/org/gnome/desktop/peripherals/trackball/",
+                                    trackball_entries, G_N_ELEMENTS (trackball_entries));
+        gsd_settings_migrate_check ("org.gnome.settings-daemon.peripherals.mouse.deprecated",
+                                    "/org/gnome/settings-daemon/peripherals/mouse/",
+                                    "org.gnome.desktop.peripherals.mouse",
+                                    "/org/gnome/desktop/peripherals/mouse/",
+                                    mouse_entries, G_N_ELEMENTS (mouse_entries));
+        gsd_settings_migrate_check ("org.gnome.settings-daemon.peripherals.touchpad.deprecated",
+                                    "/org/gnome/settings-daemon/peripherals/touchpad/",
+                                    "org.gnome.desktop.peripherals.touchpad",
+                                    "/org/gnome/desktop/peripherals/touchpad/",
+                                    touchpad_entries, G_N_ELEMENTS (touchpad_entries));
+}
+
 GsdMouseManager *
 gsd_mouse_manager_new (void)
 {
         if (manager_object != NULL) {
                 g_object_ref (manager_object);
         } else {
+                migrate_mouse_settings ();
                 manager_object = g_object_new (GSD_TYPE_MOUSE_MANAGER, NULL);
                 g_object_add_weak_pointer (manager_object,
                                            (gpointer *) &manager_object);
